From iris_simp_lang Require Import simp adequacy examples.par diaframe_simp_lang.
From diaframe.lib Require Import own_hints.
From iris.algebra Require Import lib.frac_auth numbers.
From iris.bi.lib Require Import fractional.
From iris.base_logic.lib Require Import invariants ghost_var.
From iris Require Import options.
Set Default Proof Using "Type*".

Open Scope Z.










Definition parallel_add: expr :=
  let: "r" := ref #0 in
  (FAA "r" #2) ||| (FAA "r" #2);;
  !"r".

Definition parallel_addR := frac_authR natR.
Class parallel_addG Σ :=
  ParallelAddG { parallel_add_inG :> inG Σ parallel_addR }.
Definition parallel_addΣ : gFunctors :=
  #[GFunctor parallel_addR].

Section proof.
  Context `{!simpGS Σ, !spawnG Σ, !parallel_addG Σ}.

  Definition parallel_add_inv (r : loc) (γ : gname) : iProp Σ :=
    ∃ n : nat, r ↦ #n ∗ own γ (●F n).

  (** MANUAL version for comparison *)
  Lemma parallel_add_spec_manual :
    {{{ True }}} parallel_add {{{ RET #4; True }}}.
  Proof.
    iIntros (Φ) "_ Post".
    unfold parallel_add. wp_alloc r as "Hr". wp_let.
    iMod (own_alloc (●F O ⋅ ◯F O)) as (γ) "[H●γ [H◯γ1 H◯γ2]]".
    { by apply auth_both_valid. }
    iMod (inv_alloc nroot _ (parallel_add_inv r γ) with "[Hr H●γ]") as "#Hinv".
    { iExists O. iFrame. }
    wp_apply (wp_par (λ _, own γ (◯F{1/2} 2%nat)) (λ _, own γ (◯F{1/2} 2%nat))
                with "[H◯γ1] [H◯γ2]").
    - iInv "Hinv" as (z) ">(Hr & H●γ)".
      wp_apply (wp_faa with "Hr"); iIntros "Hr".
      iMod (own_update_2 with "H●γ H◯γ1") as "[H●γ H◯γ1]". 
      { by eapply (frac_auth_update _ _ _ (z + 2)%nat 2%nat), nat_local_update. }
      iFrame. iExists _; iFrame.
      replace (Z.of_nat (z + 2)%nat) with (z + 2) by lia; done.
    - iInv "Hinv" as (z) ">(Hr & H●γ)".
      wp_apply (wp_faa with "Hr"); iIntros "Hr".
      iMod (own_update_2 with "H●γ H◯γ2") as "[H●γ H◯γ1]". 
      { by eapply (frac_auth_update _ _ _ (z + 2)%nat 2%nat), nat_local_update. }
      iFrame. iExists _; iFrame.
      replace (Z.of_nat (z + 2)%nat) with (z + 2) by lia; done.
    - iIntros (??) "[H◯γ1 H◯γ2] !>". wp_seq.
      iCombine "H◯γ1 H◯γ2" as "H◯γ".
      iInv "Hinv" as (z) ">(Hr & H●γ)".
      wp_load.
      iDestruct (own_valid_2 with "H●γ H◯γ") as %->%frac_auth_agree_L.
      iSplitL "Hr H●γ".
      { iExists  _; by iFrame. }
      by iApply "Post".
  Qed.

  (** Proof using Diaframe *)
  Lemma parallel_add_spec_autom :
    {{{ True }}} parallel_add {{{ RET #4; True }}}.
  Proof.
    unfold parallel_add. iStepsS.
    iAssert (|={⊤}=> ∃ γ, inv nroot (parallel_add_inv x0 γ) ∗ own γ (◯F O))%I 
      with "[H2]" as ">[%γ (#Hinv & H◯γ1 & H◯γ2)]".
    { iStepsS. } iIntros "!>".
    wp_apply (wp_par (λ _, own γ (◯F{1/2} 2%nat)) (λ _, own γ (◯F{1/2} 2%nat))
                with "[H◯γ1] [H◯γ2]").
    all: iStepsS.
  Qed.

End proof.
